# =====================================================================
# cgi.awk: decode CGI variables.
#
# Copyright (c) 2002,2003,2004 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# Notes:
#
# Decodes HTTP request parameters and turns them into valid rc(1)
# assignments. This script contains an AWK re-implementation and
# extension of the "uncgi" utility. See uncgi home page for more info:
# http://www.midwinter.com/~koreth/uncgi.html With respect to the
# original program, this script adds several CSA- and rc(1)- specific
# enhancements, and supports file-uploads.
# =====================================================================
# $Id: cgi.awk,v 1.2 2004/08/11 16:26:52 carlo Exp $

BEGIN {

   rc = 0

   # Safety checks.
   if (ENVIRON["TMPDIR"] !~ /^\// ||
       ENVIRON["libstem"] !~ /^\// ||
       _PID !~ /^[1-9][0-9]*$/) exit(rc=1)

   # Get file-names from env.
   o_file = ENVIRON["CSA_RPC_CMD"]
   awk_file = ENVIRON["CSA_RPC_AWK"]
   arg_file = ENVIRON["CSA_RDF_ARG"]
   put_file = ENVIRON["CSA_RPC_PUT"]

   first = 1

   print "BEGIN{\n_creat(ENVIRON[\"CSA_RPC_WWW\"],_O_TRUNC)" > awk_file

   # Supported national languages.

   delete _LANG

   _LANG["en"] = "en"
   _LANG["en_UK"] = "en_UK"
   _LANG["en_US"] = "en_US"
   _LANG["es"] = "es"
   _LANG["it"] = "it"

   # Misc. settings.
   rpcfile = map_file = "/dev/null"
   file_stem = ENVIRON["TMPDIR"] "/hexBinary." _PID

   # Max. input size, in bytes.
   rpc_maxsize = ENVIRON["CSA_RPC_MAXSIZE"]
   if (!rpc_maxsize) rpc_maxsize = 10000

   # First: decode HTTP cookie, if any.
   #
   # Note: when a cookie is set more than once in the client browser,
   # the latter returns the latest setting first. It will be up to the
   # calling rc(1) program to account for this and explicitly reference
   # the relevant rc(1) list element if appropriate.
   # Since cookie-based variables are taken into account first, their
   # values will be listed first in the resulting array, before
   # any form and URI values.
   # The order is: cookie, headers, POST, QUERY_STRING, PATH_INFO.

   if (ENVIRON["HTTP_COOKIE"] != "") {
      if (_uncgi(_hdrtocgi(ENVIRON["HTTP_COOKIE"]),o_file,1) == _EMSGSIZE) {

	 print "HTTP_COOKIE=()" > o_file		# shrink env.

	 # Exit with an error if the max. request message
	 # size is exceeded.

	 print "csaExit.fault 413" > o_file
      }
   }

   # Second: decode HTTP custom headers.
   #
   # HTTP custom headers are treated exactly the same as HTTP cookies.
   # In case of multiple settings for the same variable through
   # different means, the following order applies in the resulting rc(1)
   # list: cookie, headers, POST, QUERY_STRING, PATH_INFO. Special
   # characters passed in header values must be URI-encoded, as usual.

   for (env in ENVIRON) {
       e = env					# Save before mangling.
       if (!sub(/^HTTP_X_/,"",env)) continue
       if (env == "") continue			# re-check after mangling.

       if (_uncgi(_hdrtocgi("X-" env "=" ENVIRON[e]),o_file) == _EMSGSIZE) {

	  # Exit with an error if the max. request message
	  # size is exceeded.

	  print "csaExit.fault 413" > o_file
       }
       printf("HTTP_X_%s=()\n", env) > o_file	# shrink env.
   }

   if (ENVIRON["CONTENT_LENGTH"] > rpc_maxsize) {

      # Exit with an error if the max. input size is exceeded. This
      # is only a first check, that relies on the Content-Length HTTP
      # header. We do not control such header, although it can usually
      # be relied upon, but as a last-resort measure I check the actual
      # input size again further down, just in case.

      print "csaExit.fault 413" > o_file
    }
}

# Store the input entity body if "PUT". It can be either an HTTP or a
# CSA PUT (see the CSA front.end). In any case the entity body gets
# stored to a temporary file unchanged. Note that for a possible ending
# newline in the input not to be dropped, it is required that 'csa-escape -z'
# adds an extra trailing newline in the data itself. Such extra character
# will not affect the multipart format, due to the presence of the MIME
# part boundary in the relevant input.

ENVIRON["REQUEST_METHOD"] == "PUT" {

   if (!first) {
      if (hold) printf("\r\n") | _CMD
      else printf("\n") | _CMD
   }
   else _CMD = "csa-unescape -z > " put_file

   if (sub(/\r$/,"")) hold = 1
   else hold = 0
   printf("%s", $0) | _CMD
   first = 0
   next
}

# handle POST

ENVIRON["REQUEST_METHOD"] == "POST" {

   if (ENVIRON["CONTENT_TYPE"] == "application/x-www-form-urlencoded") {
      if (input != "") input = input "\n"
      input = input $0
      next
   }

   # otherwise assume content-type is multipart/form-data

   if (NR == 1) {

      boundary = $0

      # Safety check.
      if ($0 !~ /^--.+\r$/) exit(rc=1)	# Not a MIME boundary.

      eof = $0
      sub(/\r$/,"",eof)
      eof = eof "--"
   }
}

# Note: the actual boundary is whatever string is indicated in the
# "boundary" attribute, prepended by two hyphens "--". That is, for
# "boundary=AaB03x" the actual boundary will be "--AaB03x", and the
# ending boundary will be "--AaB03x--".

# Note that to be rigorous, I should un-escape the backslashes that
# were introduced by 'csa-escape -z' also in MIME headers, but such
# characters are not normally present in those parts of the input
# message, so I can leave without unescaping them here.

$0 == boundary {name = ""; header = $0 "\n"; binary = 0; next}

$0 == eof || $0 == eof "\r" {exit}

header != "" {
   header = header $0 "\n"
   if ($0 ~ /^\r$/) {
      binary = 0
      name = header

      if (sub(/filename="[^"]+"/,"",header)) {
	 binary = 1
	 first = 1
      }

      header = name				# restore from backup
      name = ""

      #printf("%s", header)                  # debug
      if (sub(/^.*: form-data; *name="/,"",header)) {
         sub(/".*/,"",header)
         name = header; norm = rec = 0

	 # Unlike uncgi, for names to be normalized
	 # they must begin with a dot.

         if (!binary) norm = sub(/^\./, "", name)

         # replace invalid characters with underscores (_) in names
         gsub(/[^A-Za-z0-9_]/, "_", name)

         if (name == "") exit(rc=1)	# Unexpected input format.
         #print name                         # debug
      }
      header = ""
   }
   next
}

binary {
  if (!first) {
     if (hold) printf("\r\n") | _CMD
     else printf("\n") | _CMD
  }
  else {
     if (o_tmp != "") close(_CMD)
     o_tmp = file_stem "." name ".tmp"
     _cgivar(name,o_tmp,1)
     _CMD = "csa-unescape -z > " o_tmp 
  }
  if (sub(/\r$/,"")) hold = 1
  else hold = 0
  printf("%s", $0) | _CMD
  first = 0
  next
}

header == "" {
  #print			# debug
  sub(/\r$/,"")			# Remove trailing CR's
  if (norm) {
     gsub(/\r/,"\n")		# Turn embedded CR's into NL's.
     sub(/^[ \t]+/,"")		# Strip leading white-space.
  }

  rec++				# detect multiline input (textarea)

  # Un-escape backslashes that were escaped by "csa-escape -z" (see the CSA
  # main wrapper). Binary zeroes are not unescaped here, of course, or
  # mawk(1) would break, but on the other hand they are not expected in
  # STDIN input parts other than files.

  gsub(/\134\134/, "\134")

  # append new list element only if 1st (or only) record of input,
  # otherwise append to the same element.

  if (rec == 1) _cgivar(name,$0)
  else {
     _CGIVAL[name] = _CGIVAL[name] "\n" $0
     # multi-line values can only map to scalars.
     if (rec == 2) sub(/.?$/,"0",_RDFMAP[name])
  }
}

END {

   if (_CMD != "") close(_CMD)	# Close the PUT stream if open.

   # Decode POST variables, then GET variables, and finally PATH_INFO
   # variables last. Like uncgi, treat HTTP GET and POST the same.
   # Given the decoding order, GET overrides POST, and PATH_INFO
   # overrides GET. Note: due to this process, if a generic variable is
   # assigned multiple values in the three methods, those values will
   # be listed in reverse order in the resulting rc(1) list.
   # It is recommended that a variable be assigned with only one
   # mechanism.

   # Note: if urlencoded input comes from an XML attribute value, either
   # as PATH_INFO or QUERY_STRING data, it may be XML-encoded, so I
   # need to account for that. In fact, at lead in case of a Web browser,
   # the decoding *may* by done by the browser itself, but double-
   # checking won't hurt.

   ENVIRON["QUERY_STRING"] = _xmldecode(ENVIRON["QUERY_STRING"])
   ENVIRON["PATH_INFO"] = _xmldecode(ENVIRON["PATH_INFO"])

   # Apply the site-wide remapping rules. Another good location for
   # thi file would be /etc.

   rpcappl = _pathinfo(ENVIRON["CSA_INSTALL"] "/etc/rest.map")

   # Apply the application-specific remapping rules. Note that the
   # target application name is not yet known, so we had to work it
   # out from the site-wide map first. If the latter is not available
   # then no remapping will be done at all.

   if (rpcappl != "")
   	_pathinfo(ENVIRON["libstem"] "/" rpcappl "/lib/rest.map")

   gsub(";", "/", ENVIRON["PATH_INFO"])		# see RFC-2396

   n = split(ENVIRON["PATH_INFO"], a, "/")

   if (ENVIRON["QUERY_STRING"] != "")
		ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] "&"

   # Consider class and method here only if present.
   if (n >= 3)
      ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] "0=" a[2] "." a[3] "&"

   for (i=4; i<=n; i++) {

     # Individual slash-delimited PATH_INFO tokens are treated like
     # HTTP custom headers and cookies if they contain the special
     # characters "=" or "&". Otherwise they are treated as
     # pseudo-positional arguments. Names that begin with "X-"
     # (case-insensitive) are turned into upper-case.

     if (a[i] ~ /[=&]/)
	ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] _hdrtocgi(a[i]) "&"

     else ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] i-3 "=" a[i] "&"
   }

   if (_uncgi(input,o_file) == _EMSGSIZE || \
	_uncgi(ENVIRON["QUERY_STRING"],o_file) == _EMSGSIZE) {

      # Exit with an error if the max. request message
      # size is exceeded.

      print "csaExit.fault 413" > o_file
   }

   # decode RwA payload first, if present.

   for (i in _CGIVAL)
       if (i == "CSA_RWA") _rwa2csa(_CGIVAL[i])

   # print the resulting assignments

   for (i in _CGIVAL) {

       if (i == "0" || i == "RwA") continue

       # Split multi-valued args and print RDF call statements. If an
       # associated RwA object exists, it means that it was detected
       # during a multipart POST operation (see further down). In this
       # case the parameter is expected to contain a single value in the
       # form of an absolute path to a temporary file where the binary
       # content was stored. The resulting RDF statement is thus set to
       # point at the relevant RwA attachment. Note that the resource
       # shown is a fake one, not to expose actual path names on the
       # server.

       printf("<param><rdf:Seq rdf:ID=\"call.%s\">\n",_xmlencode(i)) > arg_file

       n=split(_CGIVAL[i], a, "\001"); split(_RDFMAP[i],b,"\001")

       for (j=1; j<=n; j++) {
	 printf("<rdf:li rdf:") > arg_file
	 if (b[j]) {
	    printf("resource=\"") > arg_file

	    #if (b[j] == 1) printf("#attachment.%s\" />\n",i) > arg_file
	    #else printf("%s\" />\n",a[j]) > arg_file

	    if (b[j] == 1) {
	       # Due to attachments and b64 auto-naming, the actual
	       # variable name can only be inferred from the associated
	       # temporary file name.
	       # To prevent large inline blocks og b64 data in the <call>
	       # section of the response, inlines are represented as pseudo
	       # RwA attachments.
	       sub(/.tmp$/,"",a[j]); sub(/.*\./,"",a[j])
	       printf("#attachment.%s\" />\n",a[j]) > arg_file
	    }
	    else printf("%s\" />\n",a[j]) > arg_file
	 }
	 else printf("datatype=\"%s\">%s</rdf:li>\n",\
			_rdftype(a[j]),_xmlencode(a[j])) > arg_file
       }
       print "</rdf:Seq></param>" > arg_file

       # Handling method-calls may require data from other fields first,
       # so it needs to be done last. The drawback is that we will not
       # be able to determine the called application/method right
       # from the beginning, nor to assess in what user-level library,
       # if any, the local hook functions invar() and outvar() are stored.
       # That's why we need to store the WWW_xxx assignments in the
       # temporary AWK script 'awk_file', to be able to use them when
       # we know what method to call.

       if (i == "X_CSA_LANG") {
	  if (_LANG[_CGIVAL[i]] != "")
	     printf("CSA_LANG=%s\n",_CGIVAL[i]) > o_file
       }

       else if (i == "X_CSA_FMT") _rcassign("CSA_RPC_FMT", _CGIVAL[i], o_file)

       else if (i == "CSA_ISINDEX") _rcassign(i, _CGIVAL[i], o_file)

       # The RWA message may be big, do not fully pass it to rc(1), not
       # to bloat the shell environment. Only the first few chars are
       # assigned, for documentational purposes.

       else if (i == "CSA_RWA")
		_rcassign(i, substr(_CGIVAL[i],1,30) " ...",o_file)

       else if (_CGIVAL[i] != "") {

	  # Delay invar() invocation to when we will know what
	  # application and method have been called.
	  printf("_invar(\"WWW_%s\",%s)\n",i,_escape(_CGIVAL[i])) > awk_file
       }
   }

   # Get HTTP basic authorization data, if available.
   #
   # This header is set by the client browser in reply to a
   # "WWW-Authenticate" request from the server. It is decoded last so
   # that it will override authentications provided through other means,
   # like an HTTP cookie or an application-level token.

   _authset(ENVIRON["HTTP_AUTHORIZATION"],0,o_file)

   # Tell the application program that there are no more arguments,
   # and close the BEGIN{} block.
   print "_invar(\"_ENOMSG\"," _ENOMSG ")}" > awk_file

   # Now handle the CSA-RPC method call, if any.
   # Truncate the method name to 64 characters and extract
   # application and program name.

   if (_CGIVAL["0"] != "") {

      rpc = substr(_CGIVAL["0"],1,64)
      rpcappl = rpc; sub(/\..*$/,"",rpcappl)
      rpcname = rpc; sub(/^[^.]+\./,"",rpcname)

      if (rpcappl ~ /^[a-zA-Z0-9][-_:a-zA-Z0-9]*$/ && \
          rpcname ~ /^[a-zA-Z0-9][-_.:a-zA-Z0-9]*$/) {

	  csarc = ENVIRON["libstem"] "/" \
			rpcappl "/csa-" ENVIRON["CSA_VERSION"] ".rc"

          printf("CSA=%s\n",csarc) > o_file

	  # Switch to default profile if vesioned one not found.

	  if (_stat(csarc)) {
	     csarc = ENVIRON["libstem"] "/" rpcappl "/csa.rc"
             printf("CSA=%s\n",csarc) > o_file
	  }

	  # This test is necessary, otherwise if also the default file is
	  # missing the error msg will be embedded into the HTTP response
	  # message.

	  if (_stat(csarc)) print "csaExit.fault 404" > o_file

	  printf("CSA_PGM=($CSA_PGM '%s')\n",rpcname) > o_file

          print ". $CSA || csaExit.fault" > o_file

	  # This can be useful for generating MIME multipart responses.
	  printf("CSA_MIME_BOUNDARY=%s\n", _mimebound()) > o_file

          print "csaTrapFile $CSA_RPC_CMD $CSA_RPC_AWK $CSA_RDF_ARG $CSA_RPC_PUT " > o_file

          print "~ $CSA_DEBUG 0 || cat $CSA_RPC_CMD >[1=2]" > o_file

          print "csaLoadLib cgilib.rc || csaExit.fault" > o_file

	  # From now on, csaExit.fault() is the one provided by cgilib.rc.

	  print "csaMkTemp CSA_RPC_WWW CSA_RDF_^(TOC BODY)" > o_file

	  # Does the method-specific RDF file exist ?
	  if (!_stat(ENVIRON["libstem"] "/" rpcappl "/doc/" rpcname ".rdf")) {
	     print "CSA_RPC_IDL=" ENVIRON["libstem"] "/" rpcappl "/doc/" rpcname ".rdf" > o_file
	  }

	  # Does the application-level argument-processing AWK
	  # library exist ? Otherwise use the CSA default library.
	  # Note that the user program has no way to specify
	  # additional libraries. If they are needed, they will
	  # have to be statically included in the user-level library.
	  # If the user-level library exists, it must define at least
	  # _invar() and _outvar().

	  if (!_stat(ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk")) {
	     # Set application-specific I/O library.
	     print "CSA_RPC_IOLIB=" ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk" > o_file
	  }
	  else {
	     # Set default I/O library.
	     print "CSA_RPC_IOLIB=$CSA_INSTALL/lib/rpciolib.awk" > o_file
	  }

	  # This will write rc(1) assignments to CSA_RPC_WWW.
	  printf("csaAwk %s -- -- /dev/null || csaExit.fault 550 0003 csaAwk\n",awk_file) > o_file

	  # At this point CSA_RPC_WWW should have been written.
	  # Nope, don't run the generated code automatically, as
	  # it is up to CSA_PGM to decide when/if to load args.
	  #print ". $CSA_RPC_WWW || csaExit.fault 550 0008 $CSA_RPC_WWW" > o_file

          rpcfile = ENVIRON["libstem"] "/" rpcappl "/rpclib/" rpcname
      }
   }

   while (getline record < rpcfile > 0) print record > o_file
   if (close(rpcfile)) {

      # Requested method not found.
      print "csaExit.fault 404 0067" > o_file
   }
   else print "csaExit.ok" > o_file		# fail-safe default.
}

# =====================================================================
# End of program
# =====================================================================
