# =====================================================================
# cgilib.awk: basic CGI AWK functions for CSA.
#
# Copyright (C) 2001,2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: cgilib.awk,v 1.5 2003/05/07 09:11:26 carlo Exp $

# =====================================================================
# string isoencode(string s)
#
# Encode special HTML and form characters into the equivalent ISO
# entities. The set of characters that are ISO-encoded is:
#
#		& # < > | " ' ` \ \n \t \f SPACE
#
# Warning: '&' must be escaped first, then '#'. They need to be escaped
# to prevent the data from containing statements (both numeric and
# literal) that could be parsed by a Web server as valid SSI calls, or
# ampersand-escaped sequences that would act as formatting instructions
# to the Web browser. See http://www.apache.org/info/css-security/ for
# more info on the so-called Cross-Site Scripting vulnerabilities.
# =====================================================================

function isoencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand
   gsub(/#/, "\\&#35;", string)			# Hash mark
   gsub(/</, "\\&#60;", string)			# Open tag
   gsub(/>/, "\\&#62;", string)			# Close tag
   gsub(/\|/, "\\&#124;", string)		# Generic separator
   gsub(/"/, "\\&#34;", string)			# double quote
   gsub(/'/, "\\&#39;", string)			# single quote
   gsub(/`/, "\\&#96;", string)			# backtick
   # gsub(/\001/, "\\&#1;", string)		# rc(1)'s ^A
   gsub(/\n/, "\\&#10;", string)		# newline
   gsub(/\t/, "\\&#9;", string)			# tab
   gsub(/\f/, "\\&#12;", string)		# form-feed
   gsub(/ /, "\\&#32;", string)			# space
   gsub(/\\/, "\\&#92;", string)		# backslash

   return string
}

# =====================================================================
# string uriencode(string s)
#
# Perform URI-encoding of special characters, according to RFC 1378. The
# set of characters that are URI-encoded is:
#
#  % # & ? + SPACE = " \t \f ' \n \r $ ( ) [ ] ` < > \ / . |
#  ; ! * { } : ~ ,
#
# Note how '%' must be escaped first, and SPACE must be done after '+'.
# See http://www.apache.org/info/css-security/ for more info on the
# so-called Cross-Site Scripting vulnerabilities.
# =====================================================================

function uriencode(string) {

   # Order matters here.
   gsub(/%/, "%25", string)
   gsub(/\+/, "%2B", string)
   gsub(/  */, "+", string)

   # Order does not matter here.
   gsub(/\t/, "%09", string)
   gsub(/\f/, "%0C", string)
   gsub(/\n/, "%0A", string)
   gsub(/'/, "%27", string)
   gsub(/#/, "%23", string)
   gsub(/&/, "%26", string)
   gsub(/\?/, "%3F", string)
   gsub(/=/, "%3D", string)
   gsub(/"/, "%22", string)
   gsub(/\r/, "%0D", string)
   gsub(/\$/, "%24", string)
   gsub(/\(/, "%28", string)
   gsub(/\)/, "%29", string)
   gsub(/\[/, "%5B", string)
   gsub(/]/, "%5D", string)
   gsub(/`/, "%60", string)
   # gsub(/\001/, "%01", string)
   gsub(/</, "%3C", string)
   gsub(/>/, "%3E", string)
   gsub(/\\/, "%5C", string)
   gsub(/\//, "%2F", string)
   gsub(/\./, "%2E", string)
   gsub(/\|/, "%7C", string)
   gsub(/;/, "%3B", string)
   gsub(/!/, "%21", string)
   gsub(/\*/, "%2A", string)
   gsub(/{/, "%7B", string)
   gsub(/}/, "%7D", string)
   gsub(/:/, "%3A", string)
   gsub(/~/, "%7E", string)
   gsub(/,/, "%2C", string)

   # Add more encodings here, if necessary.

   return string
}

# =====================================================================
# string formtext(string name, string value, \
#				int size, int maxsize, boolean pass)
#
# Returns the proper HTML "text" form fragment, based on the specified
# arguments. Warning: the "value" string is not automatically
# ISO-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function. This is also to
# save processing, as the caller may want to encode the text only once
# for values that are needed in several places. If the last argument
# is "0" or NULL, then a normal input field is returned, otherwise
# the result will be a "password" type input field. If size/maxsize
# are NULL or "0" then they will be ignored.
# =====================================================================

function formtext(name,value,size,max,pass,		fragment) {

  if (pass) fragment = "<input type=password name=\"" name "\""
  else fragment = "<input type=text name=\"" name "\""

  size += 0; max += 0			# Cast to numbers.

  if (size) fragment = fragment " size=" size
  if (max) fragment = fragment " maxsize=" max

  fragment = fragment " value=\"" value "\">"

  return fragment
}

# =====================================================================
# string formarea(string name, string value, int cols, int rows)
#
# Returns the proper HTML "textarea" form fragment, based on the
# specified arguments. Warning: the "value" string is not automatically
# ISO-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function.
# =====================================================================

function formarea(name,value,cols,rows,		fragment) {

  fragment = "<textarea name=\"" name "\""

  cols += 0; rows += 0				# Cast to numbers.

  if (cols) fragment = fragment " cols=" cols
  if (rows) fragment = fragment " rows=" rows

  fragment = fragment ">" value "</textarea>"

  return fragment
}

# =====================================================================
# string formhidden(string name, string value)
#
# Returns the proper HTML "hidden" form fragment, based on the specified
# arguments. Warning: the "value" string is not automatically
# ISO-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function.
# =====================================================================

function formhidden(name,value,		fragment) {

  fragment = "<input type=hidden name=\"" name \
	"\" value=\"" value "\">"

  return fragment
}

# =====================================================================
# string formselect(array value, array label, string name,\
#		int selected, int size, boolean multiple, string empty)
#
# Returns the proper HTML "select" form fragment, based on the specified
# arguments. If value[] and label[] contain entries that need to be
# ISO-encoded, that must be done by the caller. Position "0" of either
# arrays must contain the No. of select options. If argument "empty"
# is passed, then the a trailing empty <option> element is appended to
# the resulting HTML object, with the specified label and a NULL value.
#
# Example:
#
# value[0] = 3
# value[1] = "1st cartoon"
# value[2] = "2nd cartoon"
# value[3] = "3rd cartoon"
#
# label[1] = "Donald Duck"
# label[2] = "Mickey Mouse"
# label[3] = "Uncle Scrooge"
#
# print formselect(value,label,"myselect",2)
#
#
# Result:
#
# <select name="myselect">
#  <option value="1st cartoon">Donald Duck</option>
#  <option selected value="2nd cartoon">Mickey Mouse</option>
#  <option value="3rd cartoon">Uncle Scrooge</option>
# </select>
# =====================================================================

function formselect(a1,a2,name,value,size,multi,empty,		\
					max,count,fragment) {

  count = 0

  # The No. of select entries is taken from element "0" of
  # either arrays.
  max = a1[0]
  if (!max) max = a2[0]

  value += 0; max += 0; size += 0; multi += 0	# Cast to numbers.

  if (empty != "") {
     a1[++max] = ""
     a2[max] = empty
  }

  if (!max) return NULL				# Nothing to do.

  if (name == NULL) name = "select1"		# Default name.

  fragment = "<select name=\"" name "\""

  if (size > 1) fragment = fragment " size=\"" size "\""
  if (multi) fragment = fragment " multiple"

  fragment = fragment ">\n"

  while (++count <= max) {
     fragment = fragment "<option"
     if (value == count) fragment = fragment " selected"

     fragment = fragment " value=\"" a1[count] "\">" \
		a2[count] "</option>\n"
  }

  fragment = fragment "</select>"

  return fragment
}

# =====================================================================
# string formradio(string name, string value, \
#			string label, boolean checked)
#
# Returns the proper HTML "radio" form fragment, based on the specified
# arguments. Warning: the "label" string is not automatically
# ISO-encoded. If such an encoding is desired it is up to the calling
# program to do it before calling this function.
# =====================================================================

function formradio(name,value,label,check,	fragment) {

  fragment = "<input type=radio name=\"" name "\""

  value += 0; check += 0			# Cast to numbers.

  if (!value) value = NULL			# "0" --> NULL

  if (check) fragment = fragment " checked"

  fragment = fragment " value=\"" value "\">" label

  return fragment
}

# =====================================================================
# int envtohtml([string template, string fileroot, string eval,\
#		string table, boolean trim, string tago,string tagc,\
#		string ofile])
#
# Take environment variables and use them to replace special tags in
# an HTML template file. All arguments are optional, but at least the
# 'template' parameter should be passed, otherwise the result will
# hardly be of any usefulness :-)
# =====================================================================

function envtohtml(tpl,froot,eval,tbl,trim,tago,tagc,ofile,	\
	ltago,ltagc,retago,retagc,nl,end,tlen,tag,enc,a,i,\
	input,record,j,fname) {

  i=j=0

  # Check args and set defaults.
  if (tpl == "") tpl = "/dev/null"
  if (ofile == "") ofile = STDOUT
  if (froot == "") froot = ENVIRON["CSA_ROOT"] "/html/"
  if (tago == "") tago = "$["
  if (tagc == "") tagc = "]"
  if (tago == tagc) return EINVAL # start- and end-tags must be different

  ltago = length(tago)
  ltagc = length(tagc)

  # Prepare the substitution regexp's.
  retago = escreg(tago)
  retagc = escreg(tagc)

  retago = retago "[-A-Za-z0-9_/.:]+" retagc

  # Load an extra NoSQL table if requested. This data is also loaded
  # in ENVIRON[], so it can be used to override data in there. Table
  # rows will be assigned to ENVIRON["csat1"], ENVIRON["csat2"], etc., with
  # ENVIRON["csat0"] being the No. of records in table. Any table header
  # information is ignored. The names used (one lower-case letter
  # follwed by one or more numbers), are CSA-reserved names (see the docs),
  # and the "t" suffix stands for "table". Note that splitting table records
  # into fields and unescaping NoSQL special sequences is up to whoever
  # will make use of these values, often through the user-level hook
  # outvar().

  if (tbl != "") {
     ENVIRON["csat0"] = 0
     while (getline record < tbl > 0) {
	# Skip table header.
	if (record ~ /^\001/) continue
	# Append row to ENVIRON[].
	ENVIRON["csat" ++ENVIRON["csat0"]] = record
     }

     close(tbl)
  }

  for (j in ENVIRON) {

      # Always skip rc(1) functions, confidential data
      # and other CSA-specific stuff.

      if (j ~ /(^fn_|PGPPASS$|^CSA_REST_PW|^csa[0-9]+$)/) {
	 ENVIRON[j] = "<!-- removed -->"
	 continue
      }


      # If CSA_AUTH_SSL is in effect, let's be careful
      # about including HTTP_AUTHORIZATION in templates.

      if (j == "HTTP_AUTHORIZATION" && ENVIRON["CSA_AUTH_SSL"]) {
	   if (ENVIRON["SSL_CIPHER_USEKEYSIZE"] == "" && \
					ENVIRON["HTTPS"] != "on") {
	      # This cannot be set to anything other than "",
	      # or there will be problems with htmltpl.awk.
	      ENVIRON[j] = ""
	      continue
	   }
      }

      # Call local output hook on non-CSA environment variables.
      if (j !~ /^CSA_/) ENVIRON[j] = outvar(j,ENVIRON[j])

      # Include also URI_ and ISO_ encodings in ENVIRON[] by default.
      ENVIRON["URI_" j] = uriencode(ENVIRON[j])
      ENVIRON["ISO_" j] = isoencode(ENVIRON[j])
  }

  # Tell the local hook that defs are over.
  outvar("ENOMSG",ENOMSG)

  if (eval != "") split(eval,a,",")

  # Process the HTML template file.

  while (getline input < tpl > 0) {
     nl = 1
     do {
        match(input,retago)
        if (RSTART) {
	   end = index(input, tagc)
	   tlen = end - RSTART - ltago
	   printf("%s", substr(input, 1, RSTART - 1)) > ofile
	   tag = substr(input, RSTART + ltago, tlen)

	   # Now handle encoding flags.
	   if (sub(/:m$/,NULL,tag)) {		# Base64.
	      enc = "base64"
	   }
	   else if (sub(/:X$/,NULL,tag)) {	# embeddable XML
	      enc = "xml"
	   }

	   for (i in a) {			# Is "tag" to be eval'ed ?
	       if (a[i] == tag) {
		  tag = ENVIRON[tag]		# Get tag value from env.
		  break
	       }
	   }

	   if (tag ~ /\//) {			# Catch file patterns.
	      if (tag ~ /\.\./) {		# Catch any '..' pattern.
		 printf("$[%s]", tag) > ofile	# Print tag literally.
	      }
	      else {
		 # Read external file, if it exists.
		 if (froot != NULL && tag !~ /^\//) {
		    tag = froot "/" tag
		    gsub(/\/\//, "/", tag)	# Turn any "//" into "/".
		 }

		 if (enc == "base64") {		# Base64.
		    system("exec mimencode -p " tag) > ofile
		 }
		 else if (enc == "xml") {
		    # embeddable XML, include literally
		    while (getline record < tag > 0) print record > ofile
		    close(tag)
		 }
		 else {
		    # generic text, perform HTML/XML escaping
		    while (getline record < tag > 0) {
		       gsub(/&/, "\\&amp;", record)
		       gsub(/</, "\\&lt;", record)
		       gsub(/"/, "\\&quot;", record)
		       print record > ofile
		    }
		    close(tag)
		 }
		 nl = 0
	      }
	   }
	   else {
	      if (trim) ENVIRON[tag] = strip(ENVIRON[tag])
	      printf("%s", ENVIRON[tag]) > ofile
	   }

	   input = substr(input, end + ltagc)
	}
	else {
	   printf("%s", input) > ofile
	   input = NULL		# Not to loop forever.
	}
     }  while (length(input))

     if (nl) printf("\n") > ofile
  }

  close(tpl)
}

# =====================================================================
# End of cgilib.awk
# =====================================================================
